# -*- coding: utf-8 -*-
#
# This file is part of essm.
# Copyright (C) 2017 ETH Zurich, Swiss Data Science Center.
#
# essm is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# essm is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with essm; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
"""Unsorted variables related to leaf model."""

from essm.variables import Variable
from essm.variables.units import (joule, kelvin, kilogram, meter, mole, pascal,
                                  second, watt)


class alpha_l(Variable):
    """Leaf albedo, fraction of shortwave radiation reflected by the leaf."""

    name = 'alpha_l'
    domain = 'real'
    latex_name = '\\alpha_l'


class a_s(Variable):
    """Fraction of one-sided leaf area covered by stomata.

    (1 if stomata are on one side only, 2 if they are on both sides).
    """

    name = 'a_s'
    domain = 'real'
    latex_name = 'a_s'


class a_sh(Variable):
    """Fraction of projected area exchanging sensible heat with the air."""

    name = 'a_sh'
    domain = 'real'
    latex_name = 'a_{sh}'
    default = 2.00000000000000


class C_wl(Variable):
    """Concentration of water in the leaf air space."""

    name = 'C_wl'
    unit = mole / meter ** 3
    domain = 'real'
    latex_name = 'C_{wl}'


class E_lmol(Variable):
    """Transpiration rate in molar units."""

    name = 'E_lmol'
    unit = mole / (meter ** 2 * second)
    domain = 'real'
    latex_name = 'E_{l,mol}'


class E_l(Variable):
    """Latent heat flux from leaf."""

    name = 'E_l'
    unit = joule / (meter ** 2 * second)
    domain = 'real'
    latex_name = 'E_l'


class epsilon_l(Variable):
    """Longwave emmissivity of the leaf surface."""

    name = 'epsilon_l'
    domain = 'real'
    latex_name = '\\epsilon_l'
    default = 1.00000000000000


class g_bw(Variable):
    """Boundary layer conductance to water vapour."""

    name = 'g_bw'
    unit = meter / second
    domain = 'real'
    latex_name = 'g_{bw}'


class g_bwmol(Variable):
    """Boundary layer conductance to water vapour."""

    name = 'g_bwmol'
    unit = mole / (meter ** 2 * second)
    domain = 'real'
    latex_name = 'g_{bw,mol}'


class Gr(Variable):
    """Grashof number."""

    name = 'Gr'
    domain = 'real'
    latex_name = 'N_{Gr_L}'


class g_sw(Variable):
    """Stomatal conductance to water vapour."""

    name = 'g_sw'
    unit = meter / second
    domain = 'real'
    latex_name = 'g_{sw}'


class g_swmol(Variable):
    """Stomatal conductance to water vapour."""

    name = 'g_swmol'
    unit = mole / (meter ** 2 * second)
    domain = 'real'
    latex_name = 'g_{sw,mol}'


class g_tw(Variable):
    """Total leaf conductance to water vapour."""

    name = 'g_tw'
    unit = meter / second
    domain = 'real'
    latex_name = 'g_{tw}'


class g_twmol(Variable):
    """Total leaf layer conductance to water vapour."""

    name = 'g_twmol'
    unit = mole / (meter ** 2 * second)
    domain = 'real'
    latex_name = 'g_{tw,mol}'


class h_c(Variable):
    """Average 1-sided convective heat transfer coefficient."""

    name = 'h_c'
    unit = joule / (kelvin * meter ** 2 * second)
    domain = 'real'
    latex_name = 'h_c'


class H_l(Variable):
    """Sensible heat flux from leaf."""

    name = 'H_l'
    unit = joule / (meter ** 2 * second)
    domain = 'real'
    latex_name = 'H_l'


class L_A(Variable):
    """Leaf area."""

    name = 'L_A'
    unit = meter ** 2
    domain = 'real'
    latex_name = 'L_A'


class L_l(Variable):
    """Leaf width as characteristic length scale for convection."""

    name = 'L_l'
    unit = meter
    domain = 'real'
    latex_name = 'L_l'


class P_wl(Variable):
    """Water vapour pressure inside the leaf."""

    name = 'P_wl'
    unit = pascal
    domain = 'real'
    latex_name = 'P_{wl}'


class r_bw(Variable):
    """Boundary layer resistance to water vapour, inverse of $g_{bw}$."""

    name = 'r_bw'
    unit = second / meter
    domain = 'real'
    latex_name = 'r_{bw}'


class r_sw(Variable):
    """Stomatal resistance to water vapour, inverse of $g_{sw}$."""

    name = 'r_sw'
    unit = second / meter
    domain = 'real'
    latex_name = 'r_{sw}'


class r_tw(Variable):
    """Total leaf resistance to water vapour, $r_{bv} + r_{sv}$."""

    name = 'r_tw'
    unit = second / meter
    domain = 'real'
    latex_name = 'r_{tw}'


class rho_al(Variable):
    """Density of air at the leaf surface."""

    name = 'rho_al'
    unit = kilogram / meter ** 3
    domain = 'real'
    latex_name = '\\rho_{al}'


class R_la(Variable):
    """Longwave radiation absorbed by leaf."""

    name = 'R_la'
    unit = watt / meter ** 2
    domain = 'real'
    latex_name = 'R_{la}'


class R_ll(Variable):
    """Longwave radiation away from leaf."""

    name = 'R_ll'
    unit = watt / meter ** 2
    domain = 'real'
    latex_name = 'R_{ll}'


class R_ld(Variable):
    """Downwards emitted/reflected global radiation from leaf."""

    name = 'R_ld'
    unit = watt / meter ** 2
    domain = 'real'
    latex_name = 'R_{ld}'


class R_lu(Variable):
    """Upwards emitted/reflected global radiation from leaf."""

    name = 'R_lu'
    unit = watt / meter ** 2
    domain = 'real'
    latex_name = 'R_{lu}'


class T_l(Variable):
    """Leaf temperature."""

    name = 'T_l'
    unit = kelvin
    domain = 'real'
    latex_name = 'T_l'


class T_w(Variable):
    """Radiative temperature of objects surrounding the leaf."""

    name = 'T_w'
    unit = kelvin
    domain = 'real'
    latex_name = 'T_w'


__all__ = (
    'alpha_l',
    'a_s',
    'a_sh',
    'C_wl',
    'E_lmol',
    'E_l',
    'epsilon_l',
    'g_bw',
    'g_bwmol',
    'Gr',
    'g_sw',
    'g_swmol',
    'g_tw',
    'g_twmol',
    'h_c',
    'H_l',
    'L_A',
    'L_l',
    'P_wl',
    'r_bw',
    'r_sw',
    'r_tw',
    'rho_al',
    'R_la',
    'R_ll',
    'R_ld',
    'R_lu',
    'T_l',
    'T_w',
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# ESSM documentation build configuration file, created by
# sphinx-quickstart on Tue May 16 14:29:58 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import sys

from sphinx.domains.python import PyModulelevel
from sphinx.ext.autodoc import AutoDirective, ClassDocumenter

from essm.equations._core import BaseEquation
from essm.variables._core import BaseVariable

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
sys.path.insert(0, os.path.abspath('..'))




class BaseDocumenter(ClassDocumenter):
    """Document expression definitions."""

    member_order = 1000

    _BASE_ATTRIBUTES = {
        'default',
        'domain',
        'expr',
        'latex_name',
        'unit',
    }

    @classmethod
    def can_document_member(cls, member, membername, isattr, parent):
        return isinstance(member, (BaseEquation, BaseVariable)) \
            and hasattr(member, 'definition')

    @property
    def object(self):
        return getattr(self, '_definition', None)

    @object.setter
    def object(self, value):
        if value is not None:
            setattr(self, '_definition', getattr(value, 'definition', value))

    def fiter_members(self, *args, **kwargs):
        members = super(BaseDocumenter, self).filter_members(*args, **kwargs)
        for (mname, member, isattr) in members:
            if mname in self._BASE_ATTRIBUTES:
                yield (mname, member, True)


def process_base(app, objtype, membername, member, skip, options):
    """Always document variables and equations."""
    if isinstance(member, (BaseEquation, BaseVariable)):
        return False


def setup(app):
    app.connect('autodoc-skip-member', process_base)
    AutoDirective._registry['base_expression'] = BaseDocumenter

    def get_attr(obj, value, *args, **kwargs):
        return getattr(obj.definition, value)

    AutoDirective._special_attrgetters[BaseVariable] = get_attr
    AutoDirective._special_attrgetters[BaseEquation] = get_attr


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Do not warn on external images.
suppress_warnings = ['image.nonlocal_uri', 'ref.citation']

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'matplotlib.sphinxext.plot_directive', 'sphinxcontrib.bibtex',
    'sphinx.ext.autodoc', 'sphinx.ext.autosummary', 'sphinx.ext.doctest',
    'sphinx.ext.intersphinx', 'sphinx.ext.todo', 'sphinx.ext.coverage',
    'sphinx.ext.mathjax', 'sphinx.ext.viewcode', 'sphinx.ext.githubpages'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'ESSM'
copyright = '2017, Stan Schymanski'
author = 'Stan Schymanski'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.

# Get the version string. Cannot be done with import!
g = {}
with open(os.path.join(os.path.dirname(__file__), '..', 'essm', 'version.py'),
          'rt') as fp:
    exec(fp.read(), g)
    version = g['__version__']

# The full version, including alpha/beta/rc tags.
release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    'description':
        'This package contains helpers to deal with physical variables '
        'and units.',
    'github_user': 'environmentalscience',
    'github_repo': 'essm',
    'github_button': False,
    'github_banner': True,
    'show_powered_by': False,
    'extra_nav_links': {
        'essm@GitHub': 'https://github.com/environmentalscience/essm',
        'essm@PyPI': 'https://pypi.python.org/pypi/essm/',
    }
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Custom sidebar templates, maps document names to template names.
html_sidebars = {
    '**': [
        'about.html',
        'navigation.html',
        'relations.html',
        'searchbox.html',
        'donate.html',
    ]
}

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'ESSMdoc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc, 'ESSM.tex', 'ESSM Documentation', 'Stan Schymanski',
        'manual'
    ),
]

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, 'essm', 'ESSM Documentation', [author], 1)]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc, 'ESSM', 'ESSM Documentation', author, 'ESSM',
        'One line description of project.', 'Miscellaneous'
    ),
]

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'https://docs.python.org/': None}

autoclass_content = 'both'

autodoc_default_flags = ['members', 'undoc-members', 'show-inheritance']



# -*- coding: utf-8 -*-
#
# This file is part of essm.
# Copyright (C) 2017 ETH Zurich, Swiss Data Science Center.
#
# essm is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# essm is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with essm; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
"""Variables module to deal with physical variables and units.

It allows attaching docstrings to variable names, defining their domains
(e.g. integer, real or complex), their units and LaTeX representations.
You can also provide a default value, which is particularly useful for
physical constants.

Creating variables
==================
To create custom variables, first import `Variable`:

    >>> from essm.variables import Variable

To define units, you must first import these units from the library:

    >>> from essm.variables.units import joule, kelvin, meter

Then you can define a custom variable with its name, description, domain,
latex_name, unit, and an optional default value, e.g.:

    >>> class demo_chamber_volume(Variable):
    ...    '''Volume of chamber.'''
    ...    latex_name = 'V_c'
    ...    domain = 'real'
    ...    name = 'V_c'
    ...    unit = meter ** 3
    ...    default = 1

    >>> class demo_chamber_length(Variable):
    ...    '''Length of chamber.'''
    ...    latex_name = 'L_c'
    ...    domain = 'real'
    ...    name = 'L_c'
    ...    unit = meter
    ...    default = 1

Now, `demo_chamber_volume` is displayed as `V_c` and
`demo_chamber_length` is displayed as `L_c`.
You can type `help(demo_chamber_volume)` to inspects its metadata.

`Variable.__defaults__` returns a dictionary with all variables and
their default values, `Variable.__units__` returns their units, and
`demo_chamber_volume.short_unit()` can be used to obtain the units
in short notation.

Creating equations
==================

To create custom equations, proceed similarly to above, i.e. first import
`Equation`:

    >>> from essm.equations import Equation

Then define an equation, e.g.:

    >>> class demo_eq_volume(Equation):
    ...    '''Calculate chamber volume.
    ...
    ...    Uses :math:`V_{l}` and assumes cubic shape.
    ...    '''
    ...    expr = demo_chamber_volume == demo_chamber_length ** 3

Importing variables and equations
=================================
You can import pre-defined variables and equations as e.g.:

    >>> from essm.variables.physics.thermodynamics import *
    >>> from essm.equations.physics.thermodynamics import *
"""

from __future__ import absolute_import

from sympy import E as e
from sympy import Eq, solve, sqrt



# -*- coding: utf-8 -*-
"""Test variables."""

import pytest

from essm.variables import Variable
from essm.variables.units import (derive_baseunit, derive_unit, joule,
                                  kilogram, markdown, meter, second)
from essm.variables.utils import generate_metadata_table
from sympy import Eq


class demo_variable(Variable):
    """Test variable."""

    default = 1
    unit = meter


class demo_variable1(Variable):
    """Test variable."""

    default = 1
    unit = meter


class demo_expression_variable(Variable):
    """Test expression variable."""

    expr = 2 * demo_variable


class lambda_E(Variable):
    unit = joule / kilogram


class E_lmass(Variable):
    unit = kilogram / (meter ** 2 * second)


class E_l(Variable):
    unit = joule / (meter ** 2 * second)


def test_variable_definition():
    """Test variable definition."""
    assert demo_variable.definition.default == 1
    assert demo_expression_variable.subs(Variable.__expressions__) \
        == 2 * demo_variable
    assert demo_expression_variable.definition.unit == meter


def test_local_definition():
    """Test local variable definition."""

    class local_definition(Variable):
        """Local definition."""
        unit = joule

        class local_variable(Variable):
            """Local variable."""
            unit = kilogram * meter ** 2 / second ** 2
            default = 2

        expr = 1.5 * local_variable

    assert local_definition.__doc__ == local_definition.definition.__doc__
    assert local_definition.definition.expr.subs(Variable.__defaults__) == 3
    assert local_definition.definition.unit == joule


def test_unit_check():
    """Test unit validation."""

    class valid_unit(Variable):
        expr = 3 * demo_variable
        unit = meter

    with pytest.raises(ValueError):

        class invalid_unit(Variable):
            expr = 4 * demo_variable
            unit = second


def test_symbolic():
    """Test that variables behave like symbols"""

    class l1(Variable):
        """Length"""
        unit = meter

    class l2(Variable):
        """Length"""
        unit = meter

    assert Eq(l1, -l2) is not False


def test_assumption():
    """Test that assumptions passed to Variable are stored"""
    class x(Variable):
        """Positive real variable."""

        assumptions = {'positive': True, 'real': True}

    assert x.is_positive is True


def test_derive_unit():
    """Test derive_unit from expression."""

    assert derive_baseunit(2 * lambda_E * E_lmass) \
        == kilogram / second ** 3
    assert derive_unit(2 * lambda_E * E_lmass) \
        == joule / (meter ** 2 * second)
    assert derive_unit(E_l / E_l) == 1
    assert(derive_unit(demo_variable - demo_variable1)) \
        == meter
    with pytest.raises(ValueError):
        derive_unit(lambda_E + E_lmass)

    class dimensionless(Variable):
        expr = demo_variable / demo_expression_variable

    assert dimensionless.definition.unit == 1


def test_remove_variable_from_registry():
    """Check is the variable is removed from registry."""

    class removable(Variable):
        """Should be removed."""

    with pytest.warns(UserWarning):
        del Variable[removable]

    assert removable not in Variable.__registry__

    with pytest.raises(KeyError):
        del Variable[removable]


def test_latex():
    """Test latex representaiton of variables."""

    from sympy import latex

    class lmbda(Variable):
        """Test variable."""
        latex_name = 'L'

    assert latex(lmbda) == 'L'


def test_markdown():
    """Check markdown representation of units."""
    assert markdown(kilogram * meter / second ** 2) == 'kg m s$^{-2}$'


def test_generate_metadata_table():
    """Check display of table of units."""
    assert generate_metadata_table([E_l, lambda_E]) \
        == [('Symbol', 'Name', 'Description', 'Default value', 'Units'),
            ('$\\lambda_E$', 'lambda_E', 'Latent heat of evaporation.',
            '2450000.0', 'J kg$^{-1}$'), ('$E_l$', 'E_l',
            'Latent heat flux from leaf.', '-', 'J m$^{-2}$ s$^{-1}$')]



# -*- coding: utf-8 -*-
#
# This file is part of essm.
# Copyright (C) 2017 ETH Zurich, Swiss Data Science Center.
#
# essm is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# essm is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with essm; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
"""Chamber mass and energy balance."""

from __future__ import absolute_import

from essm.variables.units import kelvin, meter, mole, pascal, second, watt

from .._core import Variable


class W_c(Variable):
    """Chamber width."""

    unit = meter


class L_c(Variable):
    """Chamber length."""

    unit = meter


class H_c(Variable):
    """Chamber height."""

    unit = meter


class V_c(Variable):
    """Chamber volume."""

    unit = meter ** 3


class n_c(Variable):
    """molar mass of gas in chamber."""

    unit = mole


class F_in_v(Variable):
    """Volumetric flow rate into chamber."""

    unit = meter ** 3 / second
    latex_name = 'F_{in,v}'


class F_in_mola(Variable):
    """Molar flow rate of dry air into chamber."""

    unit = mole / second
    latex_name = 'F_{in,mol,a}'


class F_in_molw(Variable):
    """Molar flow rate of water vapour into chamber."""

    unit = mole / second
    latex_name = 'F_{in,mol,w}'


class F_out_mola(Variable):
    """Molar flow rate of dry air out of chamber."""

    unit = mole / second
    latex_name = 'F_{out,mol,a}'


class F_out_molw(Variable):
    """Molar flow rate of water vapour out of chamber."""

    unit = mole / second
    latex_name = 'F_{out,mol,w}'


class F_out_v(Variable):
    """Volumetric flow rate out of chamber."""

    unit = meter ** 3 / second
    latex_name = 'F_{out,v}'


class T_d(Variable):
    """Dew point temperature of incoming air."""

    unit = kelvin


class T_in(Variable):
    """Temperature of incoming air."""

    unit = kelvin
    latex_name = 'T_{in}'


class T_out(Variable):
    """Temperature of outgoing air (= chamber T_a)."""

    unit = kelvin
    latex_name = 'T_{out}'


class T_room(Variable):
    """Lab air temperature."""

    unit = kelvin
    latex_name = 'T_{room}'


class P_w_in(Variable):
    """Vapour pressure of incoming air."""

    unit = pascal
    latex_name = 'P_{w,in}'


class P_w_out(Variable):
    """Vapour pressure of outgoing air."""

    unit = pascal
    latex_name = 'P_{w,out}'


class R_H_in(Variable):
    """Relative humidity of incoming air."""

    unit = 1 / 1
    latex_name = 'R_{H,in}'


class L_A(Variable):
    """Leaf area."""

    unit = meter ** 2


__all__ = (
    'W_c',
    'L_c',
    'H_c',
    'V_c',
    'n_c',
    'F_in_v',
    'F_in_mola',
    'F_in_molw',
    'F_out_mola',
    'F_out_molw',
    'F_out_v',
    'T_d',
    'T_in',
    'T_out',
    'T_room',
    'P_w_in',
    'P_w_out',
    'R_H_in',
    'L_A',
)



# -*- coding: utf-8 -*-
#
# This file is part of essm.
# Copyright (C) 2017 ETH Zurich, Swiss Data Science Center.
#
# essm is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# essm is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with essm; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
"""General and atmospheric thermodynamics variables."""

from essm.variables import Variable
from essm.variables.units import (joule, kelvin, kilogram, meter, mole, pascal,
                                  second, watt)


class alpha_a(Variable):
    """Thermal diffusivity of dry air."""

    name = 'alpha_a'
    unit = meter ** 2 / second
    domain = 'real'
    latex_name = '\\alpha_a'


class c_pa(Variable):
    """Specific heat of dry air."""

    name = 'c_pa'
    unit = joule / (kelvin * kilogram)
    domain = 'real'
    latex_name = 'c_{pa}'
    default = 1010.00000000000


class c_pamol(Variable):
    """Molar specific heat of dry air.

    https://en.wikipedia.org/wiki/Heat_capacity#Specific_heat_capacity
    """

    name = 'c_pamol'
    unit = joule / (kelvin * mole)
    domain = 'real'
    latex_name = 'c_{pa,mol}'
    default = 29.1900000000000


class c_pv(Variable):
    """Specific heat of water vapour at 300 K.

    http://www.engineeringtoolbox.com/water-vapor-d_979.html
    """

    name = 'c_pv'
    unit = joule / (kelvin * kilogram)
    domain = 'real'
    latex_name = 'c_{pv}'
    default = 1864


class C_wa(Variable):
    """Concentration of water in air."""

    name = 'C_wa'
    unit = mole / meter ** 3
    domain = 'real'
    latex_name = 'C_{wa}'


class D_va(Variable):
    """Binary diffusion coefficient of water vapour in air."""

    name = 'D_va'
    unit = meter ** 2 / second
    domain = 'real'
    latex_name = 'D_{va}'


class g(Variable):
    """Gravitational acceleration."""

    name = 'g'
    unit = meter / second ** 2
    domain = 'real'
    latex_name = 'g'
    default = 9.81000000000000


class Gr(Variable):
    """Grashof number."""

    name = 'Gr'
    domain = 'real'
    latex_name = 'N_{Gr_L}'


class h_c(Variable):
    """Average 1-sided convective heat transfer coefficient."""

    name = 'h_c'
    unit = joule / (kelvin * meter ** 2 * second)
    domain = 'real'
    latex_name = 'h_c'


class k_a(Variable):
    """Thermal conductivity of dry air."""

    name = 'k_a'
    unit = joule / (kelvin * meter * second)
    domain = 'real'
    latex_name = 'k_a'


class lambda_E(Variable):
    """Latent heat of evaporation."""

    name = 'lambda_E'
    unit = joule / kilogram
    domain = 'real'
    latex_name = '\\lambda_E'
    default = 2.45000000000000e6


class Le(Variable):
    """Lewis number."""

    name = 'Le'
    domain = 'real'
    latex_name = 'N_{Le}'


class M_air(Variable):
    """Molar mass of air.

    http://www.engineeringtoolbox.com/molecular-mass-air-d_679.html
    """

    name = 'M_air'
    unit = kilogram / mole
    domain = 'real'
    latex_name = 'M_{air}'
    default = 0.0289700000000000


class M_N2(Variable):
    """Molar mass of nitrogen."""

    name = 'M_N2'
    unit = kilogram / mole
    domain = 'real'
    latex_name = 'M_{N_2}'
    default = 0.0280000000000000


class M_O2(Variable):
    """Molar mass of oxygen."""

    name = 'M_O2'
    unit = kilogram / mole
    domain = 'real'
    latex_name = 'M_{O_2}'
    default = 0.0320000000000000


class M_w(Variable):
    """Molar mass of water."""

    name = 'M_w'
    unit = kilogram / mole
    domain = 'real'
    latex_name = 'M_w'
    default = 0.0180000000000000


class nu_a(Variable):
    """Kinematic viscosity of dry air."""

    name = 'nu_a'
    unit = meter ** 2 / second
    domain = 'real'
    latex_name = '\\nu_a'


class Nu(Variable):
    """Average Nusselt number over given length."""

    name = 'Nu'
    domain = 'real'
    latex_name = 'N_{Nu_L}'


class P_a(Variable):
    """Air pressure."""

    name = 'P_a'
    unit = pascal
    domain = 'real'
    latex_name = 'P_a'


class Pr(Variable):
    """Prandtl number (0.71 for air)."""

    name = 'Pr'
    domain = 'real'
    latex_name = 'N_{Pr}'


class P_N2(Variable):
    """Partial pressure of nitrogen."""

    name = 'P_N2'
    unit = pascal
    domain = 'real'
    latex_name = 'P_{N2}'


class P_O2(Variable):
    """Partial pressure of oxygen."""

    name = 'P_O2'
    unit = pascal
    domain = 'real'
    latex_name = 'P_{O2}'


class P_wa(Variable):
    """Water vapour pressure in the atmosphere."""

    name = 'P_wa'
    unit = pascal
    domain = 'real'
    latex_name = 'P_{wa}'


class P_was(Variable):
    """Saturation water vapour pressure at air temperature."""

    name = 'P_was'
    unit = pascal
    domain = 'real'
    latex_name = 'P_{was}'


class R_d(Variable):
    """Downwelling global radiation."""

    name = 'R_d'
    unit = watt / meter ** 2
    domain = 'real'
    latex_name = 'R_d'


class Re_c(Variable):
    """Critical Reynolds number for the onset of turbulence."""

    name = 'Re_c'
    domain = 'real'
    latex_name = 'N_{Re_c}'


class Re(Variable):
    """Average Reynolds number over given length."""

    name = 'Re'
    domain = 'real'
    latex_name = 'N_{Re_L}'


class rho_a(Variable):
    """Density of dry air."""

    name = 'rho_a'
    unit = kilogram / meter ** 3
    domain = 'real'
    latex_name = '\\rho_a'


class R_u(Variable):
    """Upwelling global radiation."""

    name = 'R_u'
    unit = watt / meter ** 2
    domain = 'real'
    latex_name = 'R_u'


class R_mol(Variable):
    """Molar gas constant."""

    name = 'R_mol'
    unit = joule / (kelvin * mole)
    domain = 'real'
    latex_name = 'R_{mol}'
    default = 8.31447200000000


class R_s(Variable):
    """Solar shortwave flux per area."""

    name = 'R_s'
    unit = joule / (meter ** 2 * second)
    domain = 'real'
    latex_name = 'R_s'


class sigm(Variable):
    """Stefan-Boltzmann constant."""

    name = 'sigm'
    unit = joule / (kelvin ** 4 * meter ** 2 * second)
    domain = 'real'
    latex_name = '\\sigma'
    default = 5.67000000000000e-8


class T0(Variable):
    """Freezing point in Kelvin."""

    name = 'T0'
    unit = kelvin
    domain = 'real'
    latex_name = 'T_0'
    default = 273.150000000000


class T_a(Variable):
    """Air temperature."""

    name = 'T_a'
    unit = kelvin
    domain = 'real'
    latex_name = 'T_a'


class v_w(Variable):
    """Wind velocity."""

    name = 'v_w'
    unit = meter / second
    domain = 'real'
    latex_name = 'v_w'


class x_N2(Variable):
    """Mole fraction of nitrogen in dry air."""

    name = 'x_N2'
    domain = 'real'
    latex_name = 'x_{N2}'
    default = 0.790000000000000


class x_O2(Variable):
    """Mole fraction of oxygen in dry air."""

    name = 'x_O2'
    domain = 'real'
    latex_name = 'x_{O2}'
    default = 0.210000000000000


__all__ = (
    'alpha_a',
    'c_pa',
    'c_pamol',
    'c_pv',
    'C_wa',
    'D_va',
    'g',
    'Gr',
    'h_c',
    'k_a',
    'lambda_E',
    'Le',
    'M_air',
    'M_N2',
    'M_O2',
    'M_w',
    'nu_a',
    'Nu',
    'P_a',
    'Pr',
    'P_N2',
    'P_O2',
    'P_wa',
    'P_was',
    'R_d',
    'Re_c',
    'Re',
    'rho_a',
    'R_u',
    'R_mol',
    'R_s',
    'sigm',
    'T0',
    'T_a',
    'v_w',
    'x_N2',
    'x_O2',
)



# -*- coding: utf-8 -*-
#
# This file is part of essm.
# Copyright (C) 2017 ETH Zurich, Swiss Data Science Center.
#
# essm is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# essm is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with essm; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
"""Leaf energy and water balance equations."""

from __future__ import division

from essm import Eq, e
from essm.equations import Equation
from essm.variables import Variable
from essm.variables.leaf.energy_water import (C_wl, E_l, E_lmol, Gr, H_l, L_l,
                                              P_wl, R_ll, T_l, T_w, a_s, a_sh,
                                              epsilon_l, g_bw, g_sw, g_tw,
                                              g_twmol, h_c, rho_al)
from essm.variables.physics.thermodynamics import (C_wa, Le, M_w, Nu, P_a,
                                                   P_wa, R_mol, R_s, Re, T_a,
                                                   c_pa, g, k_a, lambda_E,
                                                   nu_a, rho_a, sigm, v_w)
from essm.variables.units import kelvin, pascal


class eq_Rs_enbal(Equation):
    """Calculate R_s from energy balance.

    (Eq. 1 in :cite:`schymanski_leaf-scale_2017`)
    """

    expr = Eq(R_s, E_l + H_l + R_ll)


class eq_Rll(Equation):
    """R_ll as function of T_l and T_w.

    (Eq. 2 in :cite:`schymanski_leaf-scale_2017`)
    """

    expr = Eq(R_ll, (T_l ** 4 - T_w ** 4) * a_sh * epsilon_l * sigm)


class eq_Hl(Equation):
    """H_l as function of T_l.

    (Eq. 3 in :cite:`schymanski_leaf-scale_2017`)
    """

    expr = Eq(H_l, -(T_a - T_l) * a_sh * h_c)


class eq_El(Equation):
    """E_l as function of E_lmol.

    (Eq. 4 in :cite:`schymanski_leaf-scale_2017`)
    """

    expr = Eq(E_l, E_lmol * M_w * lambda_E)


class eq_Elmol(Equation):
    """E_lmol as functino of g_tw and C_wl.

    (Eq. 5 in :cite:`schymanski_leaf-scale_2017`)
    """

    expr = Eq(E_lmol, -(C_wa - C_wl) * g_tw)


class eq_gtw(Equation):
    """g_tw from g_sw and g_bw.

    (Eq. 6 in :cite:`schymanski_leaf-scale_2017`)
    """

    expr = Eq(g_tw, (1 / (1 / g_bw + 1 / g_sw)))


class eq_gbw_hc(Equation):
    """g_bw as function of h_c.

    (Eq. B2 in :cite:`schymanski_leaf-scale_2017`)
    """

    expr = Eq(g_bw, a_s * h_c / (Le ** (2 / 3) * c_pa * rho_a))


class eq_Cwl(Equation):
    """C_wl as function of P_wl and T_l.

    (Eq. B4 in :cite:`schymanski_leaf-scale_2017`)
    """

    expr = Eq(C_wl, P_wl / (R_mol * T_l))


class eq_Pwl(Equation):
    """Clausius-Clapeyron P_wl as function of T_l.

    (Eq. B3 in :cite:`hartmann_global_1994`)
    """

    class p_CC1(Variable):
        """Internal parameter of eq_Pwl."""

        name = 'p_CC1'
        unit = pascal
        domain = 'real'
        latex_name = 'p_1'
        default = 611.

    class p_CC2(Variable):
        """Internal parameter of eq_Pwl."""

        name = 'p_CC2'
        unit = kelvin
        domain = 'real'
        latex_name = 'p_2'
        default = 273.

    expr = Eq(
        P_wl, p_CC1 * e ** (-M_w * lambda_E * (1 / T_l - 1 / p_CC2) / R_mol)
    )


class eq_Elmol_conv(Equation):
    """E_lmol as function of g_twmol and P_wl.

    (Eq. B6 in :cite:`schymanski_leaf-scale_2017`)
    """

    expr = Eq(E_lmol, -(P_wa - P_wl) * g_twmol / P_a)


class eq_gtwmol_gtw(eq_Elmol.definition, eq_Cwl.definition,
                    eq_Elmol_conv.definition):
    """g_twmol as a function of g_tw.

    It uses eq_Elmol, eq_Cwl and eq_Elmol_conv.
    """

    expr = Eq(
        g_twmol, -(P_a * P_wl * T_a - P_a * P_wa * T_l) * g_tw /
        ((P_wa - P_wl) * R_mol * T_a * T_l)
    )


class eq_gtwmol_gtw_iso(eq_gtwmol_gtw.definition):
    """g_twmol as a function of g_tw at isothermal conditions."""

    expr = Eq(g_twmol, P_a * g_tw / (R_mol * T_a))


class eq_hc(Equation):
    """h_c as a function of Nu and L_l.

    (Eq. B10 in :cite:`schymanski_leaf-scale_2017`)
    """

    expr = Eq(h_c, Nu * k_a / L_l)


class eq_Re(Equation):
    """Re as a function of v_w and L_l.

    (Eq. B11 in :cite:`schymanski_leaf-scale_2017`)
    """

    expr = Eq(Re, L_l * v_w / nu_a)


class eq_Gr(Equation):
    """Gr as function of air density within and outside of leaf.

    (Eq. B12 in :cite:`schymanski_leaf-scale_2017`)
    """

    expr = Eq(Gr, L_l ** 3 * g * (rho_a - rho_al) / (nu_a ** 2 * rho_al))


__all__ = (
    'eq_Rs_enbal',
    'eq_Rll',
    'eq_Hl',
    'eq_El',
    'eq_Elmol',
    'eq_gtw',
    'eq_gbw_hc',
    'eq_Cwl',
    'eq_Pwl',
    'eq_Elmol_conv',
    'eq_gtwmol_gtw',
    'eq_gtwmol_gtw_iso',
    'eq_hc',
    'eq_Re',
    'eq_Gr',
)



# -*- coding: utf-8 -*-
#
# This file is part of essm.
# Copyright (C) 2017 ETH Zurich, Swiss Data Science Center.
#
# essm is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# essm is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with essm; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
"""Equations module to deal with physical equations.

It allows attaching docstrings to equation names (including references)
and defining internal variables.

Creating equations
==================

To create custom equations, first import `Equation` and variables needed:

    >>> from essm.equations import Equation
    >>> from essm.variables.physics.thermodynamics import *

Importing pre-defined equations
=================================
You can import pre-defined equations as e.g.:

    >>> from essm.equations.physics.thermodynamics import *
"""

from __future__ import absolute_import

from ._core import Equation

__all__ = ('Equation', )



# -*- coding: utf-8 -*-
#
# This file is part of essm.
# Copyright (C) 2017 ETH Zurich, Swiss Data Science Center.
#
# essm is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# essm is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with essm; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
"""Define unit symbols."""

import functools
import operator

import sympy.physics.units as u
from sympy import Symbol
from sympy.physics.units import Dimension, Quantity, find_unit
from sympy.physics.units.dimensions import (amount_of_substance, capacitance,
                                            charge, conductance, dimsys_SI,
                                            energy, force, frequency,
                                            inductance, luminous_intensity,
                                            magnetic_density, magnetic_flux,
                                            power, pressure, temperature, time,
                                            voltage)
from sympy.physics.units.systems import SI

candela = u.candela
coulomb = u.coulomb
farad = u.farad
gray = u.gray
henry = u.henry
joule = u.joule
katal = u.katal
kelvin = u.kelvin
kilogram = u.kilogram
lux = u.lux
meter = u.meter
mol = mole = u.mol
newton = u.newton
ohm = u.ohm
pascal = u.pascal
second = u.second
siemens = u.siemens
tesla = u.tesla
volt = u.volt
watt = u.watt
weber = u.weber

SI_BASE_DIMENSIONS = {
    Quantity.get_dimensional_expr(d): d
    for d in SI._base_units
}

SI_EXTENDED_UNITS = list(SI._base_units) + [
    kelvin, candela, lux, mol, newton, pascal, joule, watt, coulomb, volt,
    farad, ohm, siemens, weber, tesla, henry
]
SI_EXTENDED_DIMENSIONS = {
    Quantity.get_dimensional_expr(d): d
    for d in SI_EXTENDED_UNITS
}


def markdown(unit):
    """Return markdown representation of a unit."""
    from sympy.printing import StrPrinter
    from operator import itemgetter
    # displays short units (m instead of meter)
    StrPrinter._print_Quantity = lambda self, expr: str(expr.abbrev)
    if unit.is_Pow:
        item = unit.args
        return '{0}$^{{{1}}}$'.format(item[0], item[1])
    if unit.is_Mul:
        str1 = ''
        tuples = []
        allargs = unit.args
        for arg in allargs:
            if arg.is_Pow:
                args = arg.args
                tuples.append((str(args[0]), args[1]))
            if isinstance(arg, Quantity):
                tuples.append((str(arg), 1))
        tuples.sort(key=itemgetter(1), reverse=True)
        tuples.sort(key=itemgetter(0))
        for item in tuples:
            if item[1] == 1:
                str1 = str1 + ' ' + item[0]
            else:
                str1 = str1 + ' {0}$^{{{1}}}$'.format(item[0], item[1])
        return str1.strip()
    else:
        return str(unit)


def unit_symbols(expr):
    """Return unit symbols."""
    from essm.variables._core import BaseVariable, Variable

    for variable in expr.atoms(BaseVariable):
        for unit in variable.definition.unit.atoms(Quantity):
            yield unit


def derive_unit(expr, name=None):
    """Derive SI unit from an expression, omitting scale factors."""
    from essm.variables import Variable
    from essm.variables.utils import extract_variables

    dim = Variable.get_dimensional_expr(expr)
    return dim.subs(SI_EXTENDED_DIMENSIONS)


def derive_baseunit(expr, name=None):
    """Derive SI base unit from an expression, omitting scale factors."""
    from essm.variables import Variable
    from essm.variables.utils import extract_variables
    from sympy.physics.units import Dimension
    from sympy.physics.units.dimensions import dimsys_SI

    variables = extract_variables(expr)
    for var1 in variables:
        q1 = Quantity('q_' + str(var1))
        q1.set_dimension(
            Dimension(Quantity.get_dimensional_expr(var1.definition.unit))
        )
        q1.set_scale_factor(var1.definition.unit)
        expr = expr.xreplace({var1: q1})
    dim = Dimension(Quantity.get_dimensional_expr(expr))
    return functools.reduce(
        operator.mul, (
            SI_BASE_DIMENSIONS[Symbol(d)] ** p
            for d, p in dimsys_SI.get_dimensional_dependencies(dim).items()
        ), 1
    )


__all__ = (
    'derive_unit', 'markdown', 'joule', 'kelvin',
    'kilogram', 'meter', 'mole', 'pascal', 'second', 'unit_symbols', 'watt'
)



# -*- coding: utf-8 -*-
#
# This file is part of essm.
# Copyright (C) 2017 ETH Zurich, Swiss Data Science Center.
#
# essm is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# essm is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with essm; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
"""Leaf radiation balance."""

from __future__ import absolute_import

from essm.variables.units import joule, kelvin, kilogram, meter, second, watt

from .._core import Variable


class alpha_l(Variable):
    """Leaf albedo, fraction of shortwave radiation reflected by the leaf."""

    unit = watt / watt


class R_d(Variable):
    """Downwelling global radiation."""

    unit = joule / second / meter ** 2


class R_la(Variable):
    """Longwave radiation absorbed by leaf."""

    unit = joule / second / meter ** 2
    latex_name = 'R_{la}'


class R_ld(Variable):
    """Downwards emitted/reflected global radiation from leaf."""

    unit = joule / second / meter ** 2
    latex_name = 'R_{ld}'


class R_lu(Variable):
    """Upwards emitted/reflected global radiation from leaf."""

    unit = joule / second / meter ** 2
    latex_name = 'R_{lu}'


class R_u(Variable):
    """Upwelling global radiation."""

    unit = joule / second / meter ** 2


class S_a(Variable):
    """Radiation sensor above leaf reading."""

    unit = joule / second / meter ** 2


class S_b(Variable):
    """Radiation sensor below leaf reading."""

    unit = joule / second / meter ** 2


class S_s(Variable):
    """Radiation sensor beside leaf reading."""

    unit = joule / second / meter ** 2


__all__ = (
    'alpha_l',
    'R_d',
    'R_la',
    'R_ld',
    'R_lu',
    'R_u',
    'S_a',
    'S_b',
    'S_s',
)



# -*- coding: utf-8 -*-
#
# This file is part of essm.
# Copyright (C) 2017 ETH Zurich, Swiss Data Science Center.
#
# essm is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# essm is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with essm; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
"""Transform expressions."""

from __future__ import absolute_import

import ast
import inspect
import sys

from sympy.core import numbers

_Number = ast.parse('numbers.Number', mode='eval').body


def get_num_func(node):
    """Get number wrapper."""
    return _Number


def extend_globals(f_globals):
    """Extend globals."""
    f_globals.setdefault('numbers', numbers)


class Numbers(ast.NodeTransformer):
    """Change Python numbers to 'symbolic' numbers."""

    def visit_Num(self, node):
        """Rewrite int / int to Fraction(int, int)."""
        func = get_num_func(node)

        if func is None:
            return node

        return ast.copy_location(
            ast.Call(
                func=func,
                args=[ast.Str(str(node.n))],
                keywords=[],
                starargs=None,
                kwargs=None,
            ), node
        )


class ClassDef(ast.NodeVisitor):
    """Extract expression definition."""

    def __init__(self):
        """Initialize class definition."""
        self.expr = None

    def visit_Assign(self, node):
        """Find 'expr = <expr>'."""
        for target in node.targets:
            if target.id != 'expr':
                continue
            expr = ast.Expression(Numbers().visit(node.value))
            ast.fix_missing_locations(expr)
            self.expr = compile(expr, '<string>', mode='eval')
            break


def unindent(source):
    """Fix indentation."""
    counter = 0
    while source[counter] == ' ':
        counter += 1
    if counter:
        return '\n'.join(line[counter:] for line in source.split('\n'))
    return source


def build_instance_expression(instance, expr, back=1):
    """Return fixed expression."""
    from .variables._core import BaseVariable
    try:
        # Evaluate expression in the original context.
        frame = sys._getframe(back + 1)

        # Find original code and convert numbers.
        code = ast.parse(unindent(inspect.getsource(instance)))
        class_def = ClassDef()
        class_def.visit(code)

        # Include names used during number replacement.
        f_globals = frame.f_globals.copy()
        extend_globals(f_globals)

        # Include locally defined variables.
        f_locals = frame.f_locals.copy()
        for name in dir(instance):
            data = getattr(instance, name)
            if isinstance(data, BaseVariable):
                f_locals[name] = data
        expr = eval(class_def.expr, f_globals, f_locals)
    except (IOError, TypeError):  # pragma: no cover
        pass

    return expr



# -*- coding: utf-8 -*-
#
# This file is part of essm.
# Copyright (C) 2017 ETH Zurich, Swiss Data Science Center.
#
# essm is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# essm is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with essm; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
"""Chamber insulation material."""

from __future__ import absolute_import

from essm.variables.units import joule, kelvin, kilogram, meter, second

from .._core import Variable


class c_pi(Variable):
    """Heat capacity of insulation material."""

    unit = joule / kilogram / kelvin
    latex_name = 'c_{pi}'


class lambda_i(Variable):
    """Heat conductivity of insulation material."""

    unit = joule / second / meter / kelvin


class rho_i(Variable):
    """Density of insulation material."""

    unit = kilogram / meter ** 3


class L_i(Variable):
    """Thickness of insulation material."""

    unit = meter


class A_i(Variable):
    """Conducting area of insulation material."""

    unit = meter ** 2


class Q_i(Variable):
    """Heat conduction through insulation material."""

    unit = joule / second


class dT_i(Variable):
    """Temperature increment of insulation material."""

    unit = kelvin


__all__ = (
    'c_pi',
    'lambda_i',
    'rho_i',
    'L_i',
    'A_i',
    'Q_i',
    'dT_i',
)



# -*- coding: utf-8 -*-
#
# This file is part of essm.
# Copyright (C) 2017 ETH Zurich, Swiss Data Science Center.
#
# essm is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# essm is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with essm; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
"""Base classes."""

from __future__ import absolute_import

import warnings


class RegistryType(type):
    """Base registry operations."""

    def __setitem__(cls, expr, definition):
        """Register expression in registry."""
        if expr in cls.__registry__:
            warnings.warn(
                '"{0}" will be overridden by "{1}"'.format(
                    cls.__registry__[expr].__module__ + ':' +
                    cls.__registry__[expr].name,
                    definition.__module__ + ':' + str(cls),
                ),
                stacklevel=2
            )
        cls.__registry__[expr] = definition

    def __delitem__(cls, expr):
        """Remove a expr from the registry."""
        if expr in cls.__registry__:
            warnings.warn(
                '"{0}" will be unregistered.'.format(
                    cls.__registry__[expr].__module__
                ),
                stacklevel=2
            )
            del cls.__registry__[expr]
        else:
            raise KeyError(expr)



# -*- coding: utf-8 -*-
#
# This file is part of essm.
# Copyright (C) 2017 ETH Zurich, Swiss Data Science Center.
# Copyright (C) 2018 LIST (Luxembourg Institute of Science and Technology).
#
# essm is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# essm is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with essm; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
"""Core variable type."""

from __future__ import absolute_import

import warnings

import six

from sympy import Abs, Add, Basic, Derivative, Function, Mul, Pow, S, Symbol
from sympy.physics.units import Dimension, Quantity, convert_to
from sympy.physics.units.dimensions import dimsys_default, dimsys_SI

from ..bases import RegistryType
from ..transformer import build_instance_expression
from .units import derive_unit


class VariableMeta(RegistryType):
    """Variable interface."""

    def __new__(cls, name, parents, dct):
        """Build and register new variable."""
        if '__registry__' not in dct:
            unit = dct.pop('unit', S.One)
            if unit == 1:
                unit = S.One
            definition = dct.pop('expr', None)

            dct.setdefault('name', name)
            dct.setdefault('assumptions', {'real': True})
            dct.setdefault('latex_name', dct['name'])
            dct.setdefault('unit', unit)

            instance = super(VariableMeta,
                             cls).__new__(cls, name, parents, dct)

            # Variable with definition expression.
            if definition is not None:
                definition = build_instance_expression(instance, definition)
                derived_unit = derive_unit(definition, name=name)

                if unit == S.One:
                    unit = derived_unit  # only if unit is None
                instance.expr, instance.unit = definition, unit

                dim_derived = dimsys_SI.get_dimensional_dependencies(
                    Quantity.get_dimensional_expr(derived_unit)
                )
                dim_unit = dimsys_SI.get_dimensional_dependencies(
                    Quantity.get_dimensional_expr(unit)
                )
                if dim_derived != dim_unit:
                    raise ValueError(
                        'Invalid expression units {0} should be {1}'.format(
                            instance.unit, unit
                        )
                    )

            expr = BaseVariable(
                instance,
                dct['name'],
                abbrev=dct['latex_name'],
                dimension=Dimension(Quantity.get_dimensional_expr(unit)),
                scale_factor=unit or S.One,
                **dct['assumptions']
            )
            instance[expr] = instance

            # Store definition as variable expression.
            if definition is not None:
                instance.__expressions__[expr] = definition

            # Store default variable only if it is defined.
            if 'default' in dct:
                instance.__defaults__[expr] = dct['default']

            # Store unit for each variable:
            instance.__units__[expr] = unit

            return expr

        return super(VariableMeta, cls).__new__(cls, name, parents, dct)

    def __delitem__(cls, expr):
        """Remove a variable from the registry."""
        super(VariableMeta, cls).__delitem__(expr)
        for name in ('__units__', '__defaults__', '__expressions__'):
            registry = getattr(cls, name)
            if expr in registry:
                del registry[expr]


@six.add_metaclass(VariableMeta)
class Variable(object):
    """Base type for all physical variables."""

    __registry__ = {}
    __defaults__ = {}
    __units__ = {}
    __expressions__ = {}

    @staticmethod
    def get_dimensional_expr(expr):
        """Return dimensions of expression."""
        expr = Variable.check_unit(expr)
        if isinstance(expr, Mul):
            return Mul(*[Variable.get_dimensional_expr(i) for i in expr.args])
        elif isinstance(expr, Pow):
            return Variable.get_dimensional_expr(expr.base) ** expr.exp
        elif isinstance(expr, Add):
            return Variable.get_dimensional_expr(expr.args[0])
        elif isinstance(expr, Derivative):
            dim = Variable.get_dimensional_expr(expr.expr)
            for independent, count in expr.variable_count:
                dim /= Variable.get_dimensional_expr(independent) ** count
            return dim
        elif isinstance(expr, Function):
            args = [Variable.get_dimensional_expr(arg) for arg in expr.args]
            if all(i == 1 for i in args):
                return S.One
            return expr.func(*args)
        elif isinstance(expr, Quantity):
            return expr.dimension.name
        elif isinstance(expr, BaseVariable):
            return Quantity.get_dimensional_expr(expr.definition.unit)
        return S.One

    @staticmethod
    def check_unit(expr):
        """Construct postprocessor for the addition.

        Checks for dimension mismatches of the addends, thus preventing
        expressions like `meter + second` to be created.
        """
        if not expr.is_Add or expr.is_Equality:
            return expr
        deset = {
            tuple(
                sorted(
                    dimsys_default.get_dimensional_dependencies(
                        Dimension(
                            Variable.get_dimensional_expr(i)
                            if not i.is_number else 1
                        )
                    ).items()
                )
            )
            for i in expr.args
        }
        # If `deset` has more than one element, then some dimensions do not
        # match in the sum:
        if len(deset) > 1:
            raise ValueError(
                "summation of quantities of incompatible dimensions"
            )
        return expr


class BaseVariable(Symbol):
    """Physical variable."""

    def __new__(
            cls,
            definition,
            name,
            abbrev,
            dimension,
            scale_factor=S.One,
            unit_system='SI',
            **assumptions
    ):
        self = super(BaseVariable, cls).__new__(
            cls, name, abbrev=abbrev, **assumptions
        )
        # self.set_dimension(dimension, unit_system=unit_system)
        # self.set_scale_factor(scale_factor)
        self.definition = definition
        return self

    @property
    def __doc__(self):
        return self.definition.__doc__

    def _latex(self, printer):
        return self.definition.latex_name


def _Quantity_constructor_postprocessor_Add(expr):
    """Construct postprocessor for the addition.

    It checks for dimension mismatches of the addends, thus preventing
    expressions like ``meter + second`` to be created.
    """
    deset = {
        tuple(sorted(dimsys_default.get_dimensional_dependencies(Dimension(
            Quantity.get_dimensional_expr(i) if not i.is_number else 1
        )).items()))
        for i in expr.args
        if i.free_symbols == set()  # do not raise if there are symbols
        # (free symbols could contain the units corrections)
    }
    # If `deset` has more than one element, then some dimensions do not
    # match in the sum:
    if len(deset) > 1:
        raise ValueError("summation of quantities of incompatible dimensions")
    return expr


Basic._constructor_postprocessor_mapping[BaseVariable] = {
    "Add": [_Quantity_constructor_postprocessor_Add],
}

__all__ = ('Variable', )



# -*- coding: utf-8 -*-
#
# This file is part of essm.
# Copyright (C) 2017 ETH Zurich, Swiss Data Science Center.
#
# essm is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# essm is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with essm; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
"""General and atmospheric thermodynamics equations."""

from __future__ import division

from essm import Eq, sqrt
from essm.equations import Equation
from essm.variables import Variable
from essm.variables.physics.thermodynamics import (M_N2, M_O2, P_N2, P_O2,
                                                   C_wa, D_va, Le, M_w, Nu,
                                                   P_a, P_wa, Pr, R_mol, Re,
                                                   Re_c, T_a, alpha_a, k_a,
                                                   nu_a, rho_a, x_N2, x_O2)
from essm.variables.units import joule, kelvin, meter, second


class eq_Le(Equation):
    """Le as function of alpha_a and D_va.

    (Eq. B3 in :cite:`schymanski_leaf-scale_2017`)
    """

    expr = Eq(Le, alpha_a / D_va)


class eq_Cwa(Equation):
    """C_wa as a function of P_wa and T_a.

    (Eq. B9 in :cite:`schymanski_leaf-scale_2017`)
    """

    expr = Eq(C_wa, P_wa / (R_mol * T_a))


class eq_Nu_forced_all(Equation):
    """Nu as function of Re and Re_c under forced conditions.

    (Eqs. B13--B15 in :cite:`schymanski_leaf-scale_2017`)
    """

    expr = Eq(
        Nu, -1 / 1000 * Pr ** (1 / 3) * (
            37 * (Re + Re_c - 1 / 2 * abs(Re - Re_c)) ** (4 / 5) - 37 * Re **
            (4 / 5) - 664 * sqrt(Re + Re_c - 1 / 2 * abs(Re - Re_c))
        )
    )


class eq_Dva(Equation):
    """D_va as a function of air temperature.

    (Table A.3 in :cite:`monteith_principles_2007`)
    """

    class p_Dva1(Variable):
        """Internal parameter of eq_Dva."""

        name = 'p_Dva1'
        unit = meter ** 2 / (kelvin * second)
        domain = 'real'
        latex_name = 'p_1'
        default = 1.49e-07

    class p_Dva2(Variable):
        """Internal parameter of eq_Dva."""

        name = 'p_Dva2'
        unit = meter ** 2 / second
        domain = 'real'
        latex_name = 'p_2'
        default = 1.96e-05

    expr = Eq(D_va, T_a * p_Dva1 - p_Dva2)


class eq_alphaa(Equation):
    """alpha_a as a function of air temperature.

    (Table A.3 in :cite:`monteith_principles_2007`)
    """

    class p_alpha1(Variable):
        """Internal parameter of eq_alphaa."""

        name = 'p_alpha1'
        unit = meter ** 2 / (kelvin * second)
        domain = 'real'
        latex_name = 'p_1'
        default = 1.32e-07

    class p_alpha2(Variable):
        """Internal parameter of eq_alphaa."""

        name = 'p_alpha2'
        unit = meter ** 2 / second
        domain = 'real'
        latex_name = 'p_2'
        default = 1.73e-05

    expr = Eq(alpha_a, T_a * p_alpha1 - p_alpha2)


class eq_ka(Equation):
    """k_a as a function of air temperature.

    (Table A.3 in :cite:`monteith_principles_2007`)
    """

    class p_ka1(Variable):
        """Internal parameter of eq_ka."""

        name = 'p_ka1'
        unit = joule / (kelvin ** 2 * meter * second)
        domain = 'real'
        latex_name = 'p_1'
        default = 6.84e-05

    class p_ka2(Variable):
        """Internal parameter of eq_ka."""

        name = 'p_ka2'
        unit = joule / (kelvin * meter * second)
        domain = 'real'
        latex_name = 'p_2'
        default = 5.63e-03

    expr = Eq(k_a, T_a * p_ka1 + p_ka2)


class eq_nua(Equation):
    """nu_a as a function of air temperature.

    (Table A.3 in :cite:`monteith_principles_2007`)
    """

    class p_nua1(Variable):
        """Internal parameter of eq_nua."""

        name = 'p_nua1'
        unit = meter ** 2 / (kelvin * second)
        domain = 'real'
        latex_name = 'p_1'
        default = 9.e-08

    class p_nua2(Variable):
        """Internal parameter of eq_nua."""

        name = 'p_nua2'
        unit = meter ** 2 / second
        domain = 'real'
        latex_name = 'p_2'
        default = 1.13e-05

    expr = Eq(nu_a, T_a * p_nua1 - p_nua2)


class eq_rhoa_Pwa_Ta(Equation):
    """rho_a as a function of P_wa and T_a.

    (Eq. B20 in :cite:`schymanski_leaf-scale_2017`)
    """

    expr = Eq(rho_a, (M_N2 * P_N2 + M_O2 * P_O2 + M_w * P_wa) / (R_mol * T_a))


class eq_Pa(Equation):
    """Calculate air pressure.

    From partial pressures of N2, O2 and H2O, following Dalton's law of
    partial pressures.
    """

    expr = Eq(P_a, P_N2 + P_O2 + P_wa)


class eq_PN2_PO2(Equation):
    """Calculate P_N2 as a function of P_O2.

    It follows Dalton's law of partial pressures.
    """

    expr = Eq(P_N2, P_O2 * x_N2 / x_O2)


class eq_PO2(eq_Pa.definition, eq_PN2_PO2.definition):
    """Calculate P_O2 as a function of P_a, P_N2 and P_wa."""

    expr = Eq(P_O2, (P_a * x_O2 - P_wa * x_O2) / (x_N2 + x_O2))


class eq_PN2(eq_Pa.definition, eq_PN2_PO2.definition):
    """Calculate P_N2 as a function of P_a, P_O2 and P_wa."""

    expr = Eq(P_N2, (P_a * x_N2 - P_wa * x_N2) / (x_N2 + x_O2))


class eq_rhoa(eq_rhoa_Pwa_Ta.definition, eq_PN2.definition, eq_PO2.definition):
    """Calculate rho_a from T_a, P_a and P_wa."""

    expr = Eq(
        rho_a, ((M_N2 * P_a - (M_N2 - M_w) * P_wa) * x_N2 +
                (M_O2 * P_a - (M_O2 - M_w) * P_wa) * x_O2) /
        (R_mol * T_a * x_N2 + R_mol * T_a * x_O2)
    )


__all__ = (
    'eq_Le',
    'eq_Cwa',
    'eq_Nu_forced_all',
    'eq_Dva',
    'eq_alphaa',
    'eq_ka',
    'eq_nua',
    'eq_rhoa_Pwa_Ta',
    'eq_Pa',
    'eq_PN2_PO2',
    'eq_PO2',
    'eq_PN2',
    'eq_rhoa',
)



# -*- coding: utf-8 -*-
#
# This file is part of essm.
# Copyright (C) 2017 ETH Zurich, Swiss Data Science Center.
#
# essm is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# essm is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with essm; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
"""Generator for equation definitions."""

import datetime
import logging
import os
import re
from collections import defaultdict

import pkg_resources
from isort import SortImports
from yapf.yapflib.yapf_api import FormatCode

import essm

from .variables import Variable

logger = logging.getLogger()

STYLE_YAPF = pkg_resources.resource_filename('essm', 'style.yapf')

LICENSE_TPL = """# -*- coding: utf-8 -*-
#
# This file is part of essm.
# Copyright (C) {year} ETH Zurich, Swiss Data Science Center.
#
# essm is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# essm is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with essm; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
"""

EQUATION_TPL = """
class {name}({parents}):
    \"\"\"{doc}\"\"\"

    {variables}
    expr = {expr}
"""

VARIABLE_TPL = """
class {name}(Variable):
    \"\"\"{doc}\"\"\"

    name = {name!r}
    unit = {units}
    domain = {domain!r}
    latex_name = {latexname!r}
    {default}
"""

# CONSTANTS = re.compile(r'\b(e|pi)\b')
_IMPORTS = re.compile(
    r'\b({0})\b'.format(
        '|'.join(name for name in dir(essm) if not name.startswith('_'))
    )
)
"""Regular expression to find specific constants and functions."""


def _lint_content(content):
    """Automatically lint the generated code."""
    content = SortImports(file_contents=content).output
    content = FormatCode(content, style_config=STYLE_YAPF)[0]
    return content


def create_module(name, doc=None, folder=None, overwrite=False):
    """Create folder with init file."""
    name_path = (name.replace('.', os.path.sep), ) \
        if not isinstance(name, (tuple, list)) \
        else name
    folder = folder or pkg_resources.resource_filename('essm', '')
    path = os.path.join(folder, *name_path)
    try:
        os.makedirs(path)
        logger.info('Created new folder: {0}'.format(path))
    except OSError as e1:
        logger.error('Could not create new folder: {0}'.format(path))

    init_path = os.path.join(path, '__init__.py')

    if os.path.isfile(init_path):
        logger.info(
            '{0} already exists. Use `overwrite=True` to overwrite.'.
            format(init_path)
        )

    if overwrite or not os.path.isfile(init_path):
        with open(init_path, 'w') as file_out:
            file_out.write(
                LICENSE_TPL.format(year=datetime.datetime.now().year)
            )
            file_out.write('"""{0}"""\n'.format(doc))
        logger.debug('Created file {0}.'.format(init_path))

    return path


class VariableWriter(object):
    """Generate Variable definitions.

    Example:

    .. code-block:: python
        from essm._generator import VariableWriter
    """

    TPL = VARIABLE_TPL
    LICENSE_TPL = LICENSE_TPL
    default_imports = {
        'essm.variables': {'Variable'},
    }

    def __init__(self, docstring=None):
        """Initialize variable writer."""
        self.docstring = docstring
        self._imports = defaultdict(set)
        self._imports.update(**self.default_imports)
        self.vars = []

    @property
    def imports(self):
        """Yield used imports."""
        for key, values in sorted(self._imports.items()):
            yield 'from {key} import {names}'.format(
                key=key, names=', '.join(sorted(values))
            )

    def __str__(self):
        """Serialize itself to string."""
        result = ''
        if self.docstring:
            result += self.LICENSE_TPL.format(
                year=datetime.datetime.now().year
            )
            result += '"""' + self.docstring + '"""\n\n'
            result += '\n'.join(self.imports) + '\n'
        result += '\n\n'.join(
            self.TPL.format(**var).replace('^', '**') for var in self.vars
        )
        if self.docstring:
            result += '\n\n__all__ = (\n{0}\n)'.format(
                '\n'.join(
                    "    '{0}',".format(var['name']) for var in self.vars
                )
            )
            result = _lint_content(result)
        return result

    def var(
            self,
            name,
            doc='',
            units=None,
            domain='real',
            latexname=None,
            value=None
    ):
        """Add new variable."""
        if not latexname:
            latexname = name
        if value is None:
            default = ''
        else:
            default = 'default = ' + str(value)
            # Skip trailing zeroes from real numbers only
            if isinstance(value, type(0.1)):
                default = 'default = ' + value.str(skip_zeroes=True
                                                   ).replace('^', '**')
        context = {
            "name": name,
            "doc": doc,
            "units": str(units).replace('^', '**') if units else '1/1',
            "domain": domain,
            "latexname": latexname,
            "default": default
        }
        self.vars.append(context)

        # register all imports of units
        if units:
            if units != 1:
                for arg in units.args:
                    self._imports['essm.variables.units'].add(str(arg))

    def write(self, filename):
        """Serialize itself to a filename."""
        with open(filename, 'w') as out:
            out.write(str(self))


class EquationWriter(object):
    r"""Generate Equation definitions.

    Example:

    .. code-block:: python

        from essm.equations import Equation
        from essm._generator import EquationWriter
        from essm.variables.units import second, meter, kelvin
        from essm.variables.physics.thermodynamics import R_s, D_va, T_a, \
            P_a, P_wa, P_N2, P_O2
        var('p_Dva1 p_Dva2')
        writer = EquationWriter(docstring="Test.")
        writer.eq('eq_Pa', P_a == P_N2 + P_O2 + P_wa,
                  doc='Sum partial pressures to obtain total air pressure.')
        writer.eq('eq_Pwa_Pa', P_wa == P_a - P_N2 - P_O2,
                  doc='Calculate P_wa from total air pressure.',
                  parents=['eq_Pa'])
        writer.eq('eq_Dva', D_va == p_Dva1*T_a - p_Dva2,
                  doc='D_va as a function of air temperature',
                  variables = [{"name": "p_Dva1", "value": '1.49e-07',
                                "units": meter^2/second/kelvin},
                               {"name": "p_Dva2", "value": '1.96e-05',
                                "units": meter^2/second}])
        print(writer)
    """

    TPL = EQUATION_TPL
    VAR_TPL = VARIABLE_TPL
    LICENSE_TPL = LICENSE_TPL
    default_imports = {
        '__future__': {'division'},
        'essm.equations': {'Equation'}
    }
    """Set up default imports, including standard division."""

    def __init__(self, docstring=None):
        """Initialise equation writer."""
        self.docstring = docstring
        self._imports = defaultdict(set)
        self._imports.update(**self.default_imports)
        self.eqs = []

    @property
    def imports(self):
        """Yield registered imports."""
        for key, values in sorted(self._imports.items()):
            yield 'from {key} import {names}'.format(
                key=key, names=', '.join(sorted(values))
            )

    def __str__(self):
        """Return string representation."""
        result = ''
        if self.docstring:
            result += self.LICENSE_TPL.format(
                year=datetime.datetime.now().year
            )
            result += '"""' + self.docstring + '"""\n\n'
        result += '\n'.join(self.imports) + '\n'
        result += '\n'.join(
            self.TPL.format(**eq).replace('^', '**') for eq in self.eqs
        )
        result += '\n\n__all__ = (\n{0}\n)'.format(
            '\n'.join("    '{0}',".format(eq['name']) for eq in self.eqs)
        )
        reformatted_result = _lint_content(result)
        return reformatted_result

    def eq(self, name, expr, doc='', parents=None, variables=None):
        """Add new equation."""
        if parents:
            parents = ', '.join(parent + '.definition' for parent in parents)
        else:
            parents = 'Equation'

        if variables:
            for variable in variables:
                variable.setdefault('latexname', variable['name'])
                variable['doc'] = "Internal parameter of {0}.".format(name)

            # Serialize the internal variables.
            writer = VariableWriter()
            internal_variables = set()
            for variable in variables:
                writer.var(**variable)
                internal_variables.add(variable['name'])
            variables = re.sub(
                r'^',
                4 * ' ',
                str(writer),
                flags=re.MULTILINE,
            )

            # Merge all imports from variables (units, Variable).
            for key, value in writer._imports.items():
                self._imports[key] |= value
        else:
            variables = ''
            writer = None
            internal_variables = set()

        context = {
            '_variable_writer': writer,
            "name": name,
            "doc": doc,
            "expr": expr,
            "parents": parents,
            "variables": variables,
        }
        self.eqs.append(context)

        # register all imports
        for arg in expr.args:
            if str(arg) not in internal_variables and\
                    arg in Variable.__registry__:
                self._imports[Variable.__registry__[arg].__module__].add(
                    str(arg)
                )

        for match in re.finditer(_IMPORTS, str(expr)) or []:
            self._imports['essm'].add(match.group())

    def write(self, filename):
        """Serialize itself to a filename."""
        with open(filename, 'w') as out:
            out.write(str(self))



# -*- coding: utf-8 -*-
#
# This file is part of essm.
# Copyright (C) 2017 ETH Zurich, Swiss Data Science Center.
#
# essm is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# essm is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with essm; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
"""Chamber insulation."""

from __future__ import absolute_import

from essm import Eq

from ...variables.chamber.insulation import A_i, L_i, Q_i, dT_i, lambda_i
from .._core import Equation


class eq_Qi(Equation):
    """Calculate ....

    :cite:`schymanski_leaf-scale_2017`
    """

    expr = Eq(Q_i, dT_i * lambda_i * A_i / L_i)
    """Describe how you got the equation."""


__all__ = ('eq_Qi', )



# -*- coding: utf-8 -*-
#
# This file is part of essm.
# Copyright (C) 2017 ETH Zurich, Swiss Data Science Center.
# Copyright (C) 2018 LIST (Luxembourg Institute of Science and Technology).
#
# essm is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# essm is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with essm; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
"""Core equation type. Contains class definitions related to equations."""

from __future__ import absolute_import

import warnings

import six

from sympy.core.relational import Eq

from ..bases import RegistryType
from ..transformer import build_instance_expression
from ..variables import Variable
from ..variables._core import BaseVariable, Variable
from ..variables.units import unit_symbols


class EquationMeta(RegistryType):
    r"""Equation interface.

    Defines an equation with a docstring and internal variables,
    if needed.

    Example:

    .. code-block:: python

        from ..variables.units import meter, second
        class test(Equation):
            '''Test equation.'''

            class d(Variable):
                '''Internal variable.'''
                unit = meter

            class t(Variable):
                '''Internal variable.'''
                unit = second

            class v(Variable):
                '''Internal variable.'''
                unit = meter/second

            expr = v == d / t

    :raises ValueError: if the units are inconsistent.

        Example:

        .. testcode:: python

           from ..variables.units import meter, second
           class test(Equation):
               '''Test equation with inconsistent units.'''

               class d(Variable):
                   '''Internal variable.'''
                   unit = meter

               class t(Variable):
                   '''Internal variable.'''
                   unit = second

               class v(Variable):
                   '''Internal variable.'''
                   unit = meter/second

               expr = v == d * t

        Since the units of v and d*t are not the same, this returns:

        .. testoutput::

           ValueError: Invalid expression units: meter/second == meter*second
    """

    def __new__(cls, name, parents, dct):
        """Build and register new variable."""
        if '__registry__' not in dct:
            dct.setdefault('name', name)
            expr = dct.pop('expr')

            instance = super(EquationMeta,
                             cls).__new__(cls, name, parents, dct)
            expr = build_instance_expression(instance, expr)
            instance.expr = expr = BaseEquation(instance, expr)
            instance[expr] = instance

            return expr

        return super(EquationMeta, cls).__new__(cls, name, parents, dct)


@six.add_metaclass(EquationMeta)
class Equation(object):
    """Base type for all equations."""

    __registry__ = {}

    @classmethod
    def args(cls):
        """Return equation arguments from registry if exist."""
        return tuple(
            Variable.__registry__.get(arg, arg)
            for arg in cls.expr.atoms(BaseVariable)
        )


class BaseEquation(Eq):
    """Add definition and short unit."""

    def __new__(cls, definition, expr):
        if not isinstance(expr, Eq):
            return expr
        # The below raises an error if units are not consistent
        Variable.check_unit(expr.lhs + expr.rhs)
        self = super(BaseEquation, cls).__new__(cls, *expr.args)
        self.definition = definition
        return self

    @property
    def __doc__(self):
        return self.definition.__doc__

    def __add__(self, other):
        """Combite two equations."""
        if not isinstance(other, Eq):
            raise TypeError(other)

        other_definition = getattr(other, 'definition', other)
        return type(
            str(self.definition) + '_and_' + str(other_definition),
            (self.definition, other_definition),
            {'expr': Eq(self.lhs + other.lhs, self.rhs + other.rhs)},
        )

    def subs(self, *args, **kwargs):  # should mirror sympy.core.basic.subs
        """Return a new equation with subs applied to both sides.

        **Examples:**

        >>> from essm.equations.physics.thermodynamics import eq_Pa, eq_PN2_PO2
        >>> from essm.variables.physics.thermodynamics import P_N2, P_O2
        >>> eq_Pa.subs({P_N2: P_O2})
        Eq(P_a, 2*P_O2 + P_wa)
        >>> eq_Pa.subs(eq_PN2_PO2)
        Eq(P_a, P_O2*x_N2/x_O2 + P_O2 + P_wa)

        """
        if len(args) == 1:
            if isinstance(args[0], Eq):
                arg1 = {args[0].lhs: args[0].rhs}
                return Eq(self.lhs.subs(arg1, **kwargs),
                          self.rhs.subs(arg1, **kwargs))
        return Eq(
            self.lhs.subs(*args, **kwargs), self.rhs.subs(*args, **kwargs)
        )


__all__ = ('Equation', 'EquationMeta')



# -*- coding: utf-8 -*-
#
# This file is part of essm.
# Copyright (C) 2017 ETH Zurich, Swiss Data Science Center.
#
# essm is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# essm is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with essm; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
"""Variables related to leaf energy and water balance."""



# -*- coding: utf-8 -*-
#
# This file is part of essm.
# Copyright (C) 2017-2019 ETH Zurich, Swiss Data Science Center.
#
# essm is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# essm is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with essm; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
"""Environmental Science using Symbolic Math."""

import datetime
import os

from setuptools import find_packages, setup

readme = open('README.rst').read()
history = open('CHANGES.rst').read()

tests_require = [
    'check-manifest>=0.25',
    'coverage>=4.0',
    'isort>=4.2.2',
    'pydocstyle>=1.0.0',
    'pytest-cache>=1.0',
    'pytest-cov>=1.8.0',
    'pytest-pep8>=1.0.6',
    'pytest>=2.8.0',
]

extras_require = {
    'docs': [
        'Sphinx>=1.5.1,<1.8',
        'matplotlib>=1.5.1',
        'sphinxcontrib-bibtex>=0.3.5',
    ],
    'generator': ['yapf>=0.16.2', ],
    'tests':
        tests_require,
}

extras_require['all'] = []
for reqs in extras_require.values():
    extras_require['all'].extend(reqs)

setup_requires = [
    'pytest-runner>=2.6.2',
    'setuptools_scm>=3.1.0',
]

install_requires = [
    'sympy>=1.3',
    'six>=1.10.0',
]

packages = find_packages()

version_template = """\
# -*- coding: utf-8 -*-
#
# This file is part of essm.
# Copyright (C) 2017-%d ETH Zurich, Swiss Data Science Center.
#
# essm is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# essm is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with essm; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
\"\"\"Version information for ESSM.\"\"\"

__version__ = {version!r}
""" % (datetime.date.today().year, )

setup(
    name='essm',
    use_scm_version={
        'local_scheme': 'dirty-tag',
        'write_to': os.path.join('essm', 'version.py'),
        'write_to_template': version_template,
    },
    description=__doc__,
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/x-rst',
    keywords='symbolic math environmental science',
    license='GPLv2',
    author='Stan Schymanski',
    author_email='schymans@gmail.com',
    url='https://github.com/environmentalscience/essm',
    packages=packages,
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    entry_points={},
    extras_require=extras_require,
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Development Status :: 1 - Planning',
    ],
)



# -*- coding: utf-8 -*-
#
# This file is part of essm.
# Copyright (C) 2017 ETH Zurich, Swiss Data Science Center.
# Copyright (C) 2018 LIST (Luxembourg Institute of Science and Technology).
#
# essm is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# essm is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with essm; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
"""Utility function for variables."""

from essm.equations._core import BaseEquation
from essm.variables._core import BaseVariable
from sympy import Eq, preorder_traversal
from sympy.core.expr import Expr

from .units import markdown


class ListTable(list):
    """Override list class to render HTML Table in Jupyter notbook.

    Takes a 2-dimensional list of the form ``[[1, 2, 3], [4, 5, 6]],``
    and renders an HTML Table in IPython Notebook.

    Source: https://calebmadrigal.com/display-list-as-table-in-ipython-notebook

    Example:

    >>> table = ListTable()
    >>> table.append(['Column1', 'Column2'])
    >>> table.append(['1', '2'])
    >>> table
    [['Column1', 'Column2'], ['1', '2']]
    """

    def _repr_html_(self):
        html = ["<table>"]
        for row in self:
            html.append("<tr>")
            for col in row:
                html.append("<td>{0}</td>".format(col))
            html.append("</tr>")
        html.append("</table>")
        return ''.join(html)


def generate_metadata_table(variables=None, include_header=True):
    """Generate table of variables, default values and units.

    If variables not provided in list ``variables``, table will contain
    all variables in ``Variables.__registry__``.
    """
    from ._core import Variable
    table = ListTable()
    variables = variables or Variable.__registry__.keys()
    if include_header:
        table.append(
            ('Symbol', 'Name', 'Description', 'Default value', 'Units')
        )

    for variable in sorted(variables,
                           key=lambda x: x.definition.latex_name.lower()):
        symbol = '$' + variable.definition.latex_name + '$'
        name = str(variable)
        doc = variable.__doc__
        val = str(Variable.__defaults__.get(variable, '-'))

        table.append(
            (symbol, name, doc, val, markdown(variable.definition.unit))
        )
    return table


def extract_variables(expr):
    """Traverse through expression and return set of variables."""
    return {
        arg
        for arg in preorder_traversal(expr) if isinstance(arg, BaseVariable)
    }


def replace_variables(expr, variables=None):
    """Replace all base variables in expression by ``variables``."""
    if not isinstance(expr, Expr):  # stop recursion
        return expr
    symbols = {
        key: getattr(key, '_name', key)
        for key in extract_variables(expr)
    }
    variables = {
        getattr(key, '_name', key): replace_variables(value)
        for key, value in (variables or {}).items()
    }
    if isinstance(expr, BaseEquation):
        return Eq(
            expr.lhs.xreplace(symbols).xreplace(variables),
            expr.rhs.xreplace(symbols).xreplace(variables)
        )
    return expr.xreplace(symbols).xreplace(variables)


def replace_defaults(expr):
    """Replace variables in expression by their default values."""
    if hasattr(expr, 'lhs'):
        expr1 = expr.lhs
        lhs = expr1.replace(
            lambda expr1: isinstance(expr1, BaseVariable), lambda expr1: expr1.
            definition.unit * expr1.definition.default
            if hasattr(expr1.definition, 'default') else expr1
        )
        expr1 = expr.rhs
        rhs = expr1.replace(
            lambda expr1: isinstance(expr1, BaseVariable), lambda expr1: expr1.
            definition.unit * expr1.definition.default
            if hasattr(expr1.definition, 'default') else expr1
        )
        return Eq(lhs, rhs)
    else:
        expr1 = expr
        return expr1.replace(
            lambda expr1: isinstance(expr1, BaseVariable), lambda expr1: expr1.
            definition.unit * expr1.definition.default
            if hasattr(expr1.definition, 'default') else expr1
        )



# -*- coding: utf-8 -*-
"""Test equations."""

import pytest

from essm import Eq
from essm._generator import EquationWriter
from essm.equations import Equation
from essm.variables import Variable
from essm.variables.units import joule, kelvin, meter, mole, second
from essm.variables.utils import (extract_variables, replace_defaults,
                                  replace_variables)
from sympy import Derivative, S, Symbol, solve
from sympy.physics.units import Quantity, length, meter


class demo_g(Variable):
    """Test variable."""

    default = 9.8
    unit = meter / second ** 2


class demo_d(Variable):
    """Test variable."""

    unit = meter


class demo_d1(Variable):
    """Test variable."""

    unit = meter


class demo_v(Variable):
    """Test variable."""

    unit = meter / second


class demo_fall(Equation):
    """Test equation."""

    class t(Variable):
        unit = second

    expr = Eq(demo_d, S(1) / S(2) * demo_g * t ** 2)


def test_equation():
    """Test variable definition."""
    assert demo_fall.__doc__ == demo_fall.definition.__doc__


def test_units():
    """Check units during definition."""
    with pytest.raises(ValueError):

        class invalid_units(Equation):

            class x(Variable):
                unit = meter

            expr = Eq(demo_g, x)


def test_units_derivative():
    """Check units of derivative."""

    class valid_units(Equation):

        expr = Eq(demo_v, Derivative(demo_d, demo_fall.definition.t))

    with pytest.raises(ValueError):

        class invalid_units_derivative(Equation):

            expr = Eq(demo_g, Derivative(demo_d, demo_fall.definition.t))


def test_integral():
    """Test that variables can be used as integration symbols."""
    from sympy import integrate

    assert demo_g * demo_fall.definition.t ** S(3) / S(6) == integrate(
        demo_fall.rhs, demo_fall.definition.t
    )


def test_args():
    """Test defined args."""
    assert set(demo_fall.definition.args()) == {
        demo_g.definition,
        demo_d.definition,
        demo_fall.definition.t.definition,
    }


def test_variable_extraction():
    """Test extract variables from expression."""
    expr = demo_fall.rhs
    assert extract_variables(expr) == {demo_g, demo_fall.definition.t}


def test_variable_replacement():
    """Test replace variables by values and symbols in expression."""
    expr = demo_fall
    vdict = Variable.__defaults__.copy()
    vdict[Symbol('x')] = 1
    assert replace_variables(expr, vdict) == \
        Eq(demo_d, 4.9 * demo_fall.definition.t ** 2)


def test_variable_defaults():
    """Test replace variables in expression by their default values."""
    expr = demo_fall
    assert replace_defaults(expr) == \
        Eq(demo_d, 4.9*demo_fall.definition.t**2*meter/second**2)


def test_substitution():
    """Test if .subs() method gives the same result as replace_variables."""
    assert demo_fall.subs(demo_d, demo_d1) == \
        replace_variables(demo_fall, {demo_d: demo_d1})


def test_unit_check():
    """Check unit test involving temperature."""

    class combined_units(Equation):

        class x_mol(Variable):
            unit = joule / mole / kelvin

        class x_J(Variable):
            unit = joule

        class x_K(Variable):
            unit = kelvin

        class x_M(Variable):
            unit = mole

        expr = Eq(x_mol, x_J / x_M / x_K)


def test_addition():
    """Check addition of equations."""

    class x(Variable):
        unit = kelvin

    class y(Variable):
        unit = kelvin

    class z(Variable):
        unit = kelvin

    class first(Equation):
        expr = Eq(x, y)

    class second(Equation):
        expr = Eq(y, z)

    assert solve(first + second)[0][x] == z
    assert solve(first + second + first + second)[0][x] == z


def test_check_unit():
    """Make sure that check_unit works with terms that are not equations."""

    assert Variable.check_unit(demo_fall.rhs) == \
        S(1) / S(2) * demo_g * demo_fall.definition.t ** 2


def test_double_registration():
    """Check double registration warning."""

    class demo_double(Equation):
        """First."""

        expr = Eq(demo_g, demo_g)

    assert Equation.__registry__[demo_double].__doc__ == 'First.'

    with pytest.warns(UserWarning):

        class demo_double(Equation):  # ignore: W0232
            """Second."""

            expr = Eq(demo_g, demo_g)

    assert Equation.__registry__[demo_double].__doc__ == 'Second.'


def test_solve():
    """Check that equation solving works"""

    demo_d2 = type('demo_d2', (Variable, ), {'unit': meter})
    assert solve(
        10 ** 6 * demo_d1 - 0.031e6 * demo_d + 0.168 * demo_d2, demo_d1
    ) == [0.031 * demo_d - 1.68e-7 * demo_d2]


@pytest.mark.skip(reason="needs rewrite for SymPy")
def test_equation_writer(tmpdir):
    """EquationWriter creates importable file with internal variables."""
    from sympy import var
    g = {}
    d = var('d')
    t = var('t')
    writer_td = EquationWriter(docstring='Test of Equation_writer.')
    writer_td.eq(
        'demo_fall',
        Eq(demo_g, d / t ** 2),
        doc='Test equation.\n\n    (Some reference)\n    ',
        variables=[{
            "name": "d",
            "value": '0.9',
            "units": meter,
            "latexname": 'p_1'
        }, {
            "name": "t",
            "units": second,
            "latexname": 'p_2'
        }]
    )
    eq_file = tmpdir.mkdir('test').join('test_equations.py')
    writer_td.write(eq_file.strpath)
    execfile(eq_file.strpath, g)
    assert g['demo_fall'].definition.d.definition.default == 0.9


@pytest.mark.skip(reason="needs rewrite for SymPy")
def test_equation_writer_linebreaks(tmpdir):
    """EquationWriter breaks long import lines."""
    from essm.variables.physics.thermodynamics import alpha_a, D_va, P_wa, \
        R_mol, T_a, M_O2, P_O2, P_N2, M_N2, M_w, Le, C_wa, rho_a, P_a

    writer_td = EquationWriter(docstring='Test of Equation_writer.')
    writer_td.eq(
        'eq_Le',
        Eq(Le, alpha_a / D_va),
        doc='Le as function of alpha_a and D_va.'
    )
    writer_td.eq(
        'eq_Cwa',
        Eq(C_wa, P_wa / (R_mol * T_a)),
        doc='C_wa as a function of P_wa and T_a.'
    )
    writer_td.eq(
        'eq_rhoa_Pwa_Ta',
        Eq(rho_a, (M_w * P_wa + M_N2 * P_N2 + M_O2 * P_O2) / (R_mol * T_a)),
        doc='rho_a as a function of P_wa and T_a.'
    )
    writer_td.eq(
        'eq_Pa',
        Eq(P_a, P_N2 + P_O2 + P_wa),
        doc='Calculate air pressure from partial pressures.'
    )
    writer_td.eq(
        'eq_PN2_PO2',
        Eq(P_N2, x_N2 / x_O2 * P_O2),
        doc='Calculate P_N2 as a function of P_O2'
    )

    eq_file = tmpdir.mkdir('test').join('test_equations.py')
    writer_td.write(eq_file.strpath)
    with open(eq_file.strpath) as outfile:
        maxlinelength = max([len(line) for line in outfile.readlines()])
    assert maxlinelength < 80
